package com.terraformersmc.terraform.block;

import java.util.function.Supplier;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1304;
import net.minecraft.class_1657;
import net.minecraft.class_1750;
import net.minecraft.class_1766;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2410;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_2689;
import net.minecraft.class_2754;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3542;
import net.minecraft.class_3620;
import net.minecraft.class_3965;

/**
 * A log block that has 4 different corners that combine together to form a huge and continuous 2x2 log.
 * Used for the mega variants of Redwood, Fir, etc
 */
public class QuarterLogBlock extends class_2410 {
	public static final class_2754<BarkSide> BARK_SIDE = class_2754.method_11850("bark_side", BarkSide.class);
	private final Supplier<class_2248> stripped;

	public QuarterLogBlock(Supplier<class_2248> stripped, class_3620 color, class_2248.class_2251 settings) {
		super(color, settings);

		this.stripped = stripped;
	}

	@Override
	protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
		super.method_9515(builder);

		builder.method_11667(BARK_SIDE);
	}

	@Override
	public class_2680 method_9605(class_1750 context) {
		class_243 pos = context.method_17698();
		class_2338 blockPos = context.method_8037();

		float hitX = (float) (pos.method_10216() - blockPos.method_10263());
		float hitY = (float) (pos.method_10214() - blockPos.method_10264());
		float hitZ = (float) (pos.method_10215() - blockPos.method_10260());

		BarkSide side = BarkSide.fromHit(context.method_8038().method_10166(), hitX, hitY, hitZ);

		return super.method_9605(context).method_11657(BARK_SIDE, side);
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_1269 method_9534(class_2680 state, class_1937 world, class_2338 pos, class_1657 player, class_1268 hand, class_3965 hit) {
		class_1799 heldStack = player.method_6118(hand == class_1268.field_5808 ? class_1304.field_6173 : class_1304.field_6171);

		if(heldStack.method_7960()) {
			return class_1269.field_5814;
		}

		class_1792 held = heldStack.method_7909();
		if(!(held instanceof class_1766)) {
			return class_1269.field_5814;
		}

		class_1766 tool = (class_1766) held;

		if(stripped != null && (tool.method_7856(state) || tool.method_7865(heldStack, state) > 1.0F)) {
			world.method_8396(player, pos, class_3417.field_14675, class_3419.field_15245, 1.0F, 1.0F);

			if(!world.field_9236) {
				class_2680 target = stripped.get().method_9564()
						.method_11657(QuarterLogBlock.field_11459, state.method_11654(QuarterLogBlock.field_11459))
						.method_11657(QuarterLogBlock.BARK_SIDE, state.method_11654(QuarterLogBlock.BARK_SIDE));

				world.method_8501(pos, target);

				heldStack.method_7956(1, player, consumedPlayer -> consumedPlayer.method_20236(hand));
			}

			return class_1269.field_5812;
		}

		return class_1269.field_5812;
	}

	/**
	 * Represents the two sides of the log the bark appears on. SOUTHEAST for example indicates that the South and East
	 * sides have bark applied, while the North and West sides are bare.
	 */
	public enum BarkSide implements class_3542 {
		SOUTHWEST("southwest"),
		NORTHWEST("northwest"),
		NORTHEAST("northeast"),
		SOUTHEAST("southeast");

		final String name;

		BarkSide(String name) {
			this.name = name;
		}

		/**
		 * Creates a new BarkSide from the fractional hit vectors, which can be derived in the manner seen above in
		 * getPlacementState. The placement logic is to ensure that the side the user clicks on is the farthest from
		 * the bark sides, providing a consistent placing experience.
		 *
		 * @param axis The axis of the log, used to determine how the hit vectors are read.
		 * @param hitX The fractional hit on the X axis, from 0.0 to 1.0.
		 * @param hitY The fractional hit on the Y axis, from 0.0 to 1.0.
		 * @param hitZ The fractional hit on the Z axis, from 0.0 to 1.0.
		 * @return A bark side according to the placement logic
		 */
		public static BarkSide fromHit(class_2350.class_2351 axis, float hitX, float hitY, float hitZ) {
			boolean hitEast;
			boolean hitSouth;

			switch (axis) {
				case field_11052:
					hitEast = hitX >= 0.5;
					hitSouth = hitZ >= 0.5;
					break;
				case field_11048:
					hitEast = hitY <= 0.5;
					hitSouth = hitZ >= 0.5;
					break;
				default:
					hitEast = hitX >= 0.5;
					hitSouth = hitY >= 0.5;
					break;
			}

			// Logic of placement: The quadrant the player clicks on should be the one farthest from the bark sides.
			return BarkSide.fromHalves(!hitEast, !hitSouth);
		}

		/**
		 * Creates a bark side from the east and south components.
		 *
		 * @param east True for east, false for west
		 * @param south True for south, false for north
		 * @return A BarkSide that is the composite of the two halves: (true, true) -> SOUTHEAST, for example.
		 */
		public static BarkSide fromHalves(boolean east, boolean south) {
			if (east) {
				if (south) {
					return BarkSide.SOUTHEAST;
				} else {
					return BarkSide.NORTHEAST;
				}
			} else {
				if (south) {
					return BarkSide.SOUTHWEST;
				} else {
					return BarkSide.NORTHWEST;
				}
			}
		}

		@Override
		public String toString() {
			return this.name;
		}

		@Override
		public String method_15434() {
			return this.name;
		}
	}
}
